"""
Integration with alpa <-> upstream repo.
"""

import shutil
import subprocess
from pathlib import Path
from typing import Optional

import click
import requests
from alpa.config import MetadataConfig
from click import ClickException
from specfile import Specfile

from alpa.repository.branch import LocalRepoBranch


class UpstreamIntegration(LocalRepoBranch):
    def __init__(self, repo_path: Path) -> None:
        super().__init__(repo_path)
        self.metadata = MetadataConfig.get_config(repo_path)
        self.specfile = Specfile(Path(repo_path / f"{self.package}.spec"))
        self.name_version = (
            f"{self.specfile.expanded_name}-{self.specfile.expanded_version}"
        )
        self.nvr = f"{self.name_version}-{self.specfile.expanded_release}"

    @staticmethod
    def _find_srpm_file_from_mock_build(srpm_result_dir: Path) -> Optional[Path]:
        for file_path in srpm_result_dir.iterdir():
            if file_path.name.endswith(".src.rpm"):
                return file_path

        return None

    @staticmethod
    def _run_mock_command(cmd: list[str]) -> int:
        click.echo(f"Executing command {' '.join(cmd)}")
        return subprocess.run(cmd).returncode

    def _mock_build_one_chroot(
        self, chroot: str, result_dir: Path, source_file_name: str
    ) -> int:
        srpm_result_dir = result_dir / "srpm" / chroot
        retval = UpstreamIntegration._run_mock_command(
            [
                "mock",
                "-r",
                chroot,
                "--buildsrpm",
                "--spec",
                self.specfile.path.name,
                "--sources",
                f"{source_file_name}.tar.gz",
                "--resultdir",
                str(srpm_result_dir),
            ]
        )
        if retval != 0:
            return retval

        rpm_result_dir = result_dir / "build_results" / chroot
        srpm_path = UpstreamIntegration._find_srpm_file_from_mock_build(srpm_result_dir)
        if srpm_path is None:
            return 1

        return UpstreamIntegration._run_mock_command(
            ["mock", "-r", chroot, "--resultdir", str(rpm_result_dir), str(srpm_path)]
        )

    def _prepare_mock_result_dir(self, chroots: list[str]) -> Path:
        mock_results_path = self.repo_path / "mock_results"
        if mock_results_path.is_dir():
            # delete previous mock build
            shutil.rmtree(mock_results_path)

        for chroot in chroots:
            srpm_chroot_path = mock_results_path / "srpm" / chroot
            srpm_chroot_path.mkdir(parents=True)
            build_chroot_path = mock_results_path / "build_results" / chroot
            build_chroot_path.mkdir(parents=True)

        return mock_results_path

    @staticmethod
    def download_upstream_source(upstream_source_url: str, name_version: str) -> None:
        resp = requests.get(upstream_source_url, allow_redirects=True)
        if not resp.ok:
            raise ConnectionError(
                f"Couldn't download source from {upstream_source_url}. "
                f"Reason: {resp.reason}"
            )

        with open(f"{name_version}.tar.gz", "wb") as archive:
            archive.write(resp.content)

    def mockbuild(self, chroots: list[str]) -> None:
        with self.specfile.sources() as sources:
            source0 = min(sources, key=lambda src: src.number)

        source_file_name = source0.expanded_filename.rstrip(".tar.gz")
        self.download_upstream_source(source0.expanded_location, source_file_name)
        root_mock_result_dir = self._prepare_mock_result_dir(chroots)

        for chroot in chroots:
            retval = self._mock_build_one_chroot(
                chroot, root_mock_result_dir, source_file_name
            )
            if retval != 0:
                raise ClickException(f"Mock returned non-zero value: {retval}")



"""
Set of commands that helps with integration of Alpa
repository.
"""
import logging
import subprocess
from abc import ABC, abstractmethod
from os import getcwd
from pathlib import Path
from typing import Optional, Iterable
from urllib.parse import urlparse

from click import UsageError, ClickException
import click

from alpa.constants import (
    ALPA_FEAT_BRANCH,
    ALPA_FEAT_BRANCH_PREFIX,
    ORIGIN_NAME,
    UPSTREAM_NAME,
)
from alpa.gh import GithubAPI, GithubRepo
from alpa.git import GitCMD
from alpa.messages import (
    CLONED_REPO_IS_NOT_FORK,
    NOT_IN_PREDEFINED_STATE,
    CLONED_REPO_IS_FORK,
    NO_PERMISSION_FOR_ALPA_REPO,
)


logger = logging.getLogger(__name__)


class LocalRepo(ABC):
    def __init__(self, repo_path: Path) -> None:
        self.repo_path = repo_path
        self.git = GitCMD(str(self.repo_path))
        self.git_cmd = self.git.git_cmd

        if not self._is_repo_in_predefined_state():
            raise ClickException(NOT_IN_PREDEFINED_STATE)

        # do it after predefined state is checked since this depends on it
        # (see comment in the _should_be_fork)
        self.remote_name = UPSTREAM_NAME if self._should_be_fork() else ORIGIN_NAME
        logger.debug(f"Remote name for this repository is set to {self.remote_name}")

        # lazy properties
        self._namespace: Optional[str] = None
        self._repo_name: Optional[str] = None

    @abstractmethod
    def get_packages(self, regex: str) -> list[str]:
        pass

    @abstractmethod
    def switch_to_package(self, package: str) -> None:
        pass

    @abstractmethod
    def _ensure_feature_branch(self) -> None:
        pass

    @abstractmethod
    def get_history_of_package(self, package: str) -> str:
        pass

    @property
    @abstractmethod
    def package(self) -> str:
        pass

    @abstractmethod
    def create_packit_config(self, override: bool) -> bool:
        pass

    @property
    def branch(self) -> str:
        return self.git_cmd(["rev-parse", "--abbrev-ref", "HEAD"]).stdout

    @staticmethod
    def get_feat_branch_of_package(package: str) -> str:
        return ALPA_FEAT_BRANCH.format(pkgname=package)

    @property
    def feat_branch(self) -> str:
        return ALPA_FEAT_BRANCH.format(pkgname=self.package)

    @staticmethod
    def _get_relevant_remote_refs(remote_refs: Iterable[str]) -> list[str]:
        relevant_refs = []
        for ref in remote_refs:
            parsed_ref = ref.split("/")[-1]
            if not parsed_ref.startswith(ALPA_FEAT_BRANCH_PREFIX):
                logger.debug(f"Relevant remote ref found: {parsed_ref}")
                relevant_refs.append(parsed_ref)

        return relevant_refs

    def get_remote_branches(self, remote: str) -> list[str]:
        lines = [
            line.strip()
            for line in self.git_cmd(
                ["remote", "--verbose", "show", remote]
            ).stdout.split("\n")
        ]
        # TODO: do a better job
        remote_branch_line = ["Remote branch:", "Remote branches:"]
        start = -1
        for line_to_match in remote_branch_line:
            if line_to_match in lines:
                start = lines.index(line_to_match)
                break

        if start == -1:
            return []

        # TODO: do a better job
        possible_start_of_local_stuff = [
            "Local branch configured for 'git pull':",
            "Local branches configured for 'git pull':",
            "Local ref configured for 'git push':",
            "Local refs configured for 'git push':",
        ]
        end = -1
        for line_to_match in possible_start_of_local_stuff:
            if line_to_match in lines:
                end = lines.index(line_to_match)
                break

        if end == -1:
            end = len(lines)

        lines_with_remote_branches = lines[start + 1 : end]
        return [line.split()[0] for line in lines_with_remote_branches]

    @property
    def remotes(self) -> set[str]:
        return set(self.git_cmd(["remote"]).stdout.split())

    def _is_repo_in_predefined_state(self) -> bool:
        return self.remotes == {ORIGIN_NAME, UPSTREAM_NAME} or self.remotes == {
            ORIGIN_NAME
        }

    def _should_be_fork(self) -> bool:
        # if repo is prepared via alpa-cli, fork should have 2 remotes and non-fork 1
        return len(self.remotes) > 1

    @property
    def untracked_files(self) -> list[str]:
        return self.git_cmd(["ls-files", "-o", "--exclude-standard"]).stdout.split()

    def _get_dirty_files(self, staged: bool) -> list[str]:
        result = []
        status_output = self.git_cmd(["status", "--porcelain=1"]).stdout
        if not status_output:
            return []

        for line in status_output.split("\n"):
            if line == "":
                # some empty line in the output
                continue

            file = line.split()[-1]
            if line.startswith("MM"):
                result.append(file)
                continue

            if not staged and line.startswith(" M"):
                result.append(file)
                continue

            if staged and line.startswith("M "):
                result.append(file)
                continue

        return result

    def is_dirty(self) -> bool:
        return self.git_cmd(["status", "--porcelain"]).stdout != ""

    @property
    def modified_files(self) -> list[str]:
        return self._get_dirty_files(False)

    @property
    def files_to_be_committed(self) -> list[str]:
        return self._get_dirty_files(True)

    @property
    def namespace(self) -> str:
        if self._namespace is not None:
            return self._namespace

        self._namespace = self.full_reponame().split("/")[0]
        logger.debug(f"Setting namespace to {self._namespace}")
        return self._namespace

    @property
    def repo_name(self) -> str:
        if self._repo_name is not None:
            return self._repo_name

        self._repo_name = self.full_reponame().split("/")[1]
        logger.debug(f"Setting repo name to {self._repo_name}")
        return self._repo_name

    @staticmethod
    def _format_files_to_status(files: list[str], msg: str) -> str:
        if not files:
            return ""

        output = msg + "\n"
        output += "\n".join(files)
        return output + "\n"

    def get_status_output(self) -> str:
        output = self._format_files_to_status(
            self.files_to_be_committed, "Files to commit:"
        )
        output += self._format_files_to_status(self.modified_files, "Modified files:")
        output += self._format_files_to_status(self.untracked_files, "Untracked files:")
        return output

    def branch_exists(self, branch: str) -> bool:
        branches = self.git_cmd(["branch"]).stdout.split()
        for ref in branches:
            if ref.strip() == branch:
                return True

        logger.debug(f"Branch {branch} does not exist in {branches}")
        return False

    def get_history_of_branch(self, branch: str, params: list[str]) -> str:
        return self.git_cmd(["log", "--decorate", "--graph"] + params + [branch]).stdout

    def commit(self, message: str, pre_commit: bool) -> bool:
        if pre_commit:
            ret = subprocess.run(["pre-commit", "run", "--all-files"])
            if ret.returncode != 0:
                return False

        self._ensure_feature_branch()
        if message:
            self.git_cmd(["commit", "-m", message])
        else:
            self.git_cmd(["commit"])

        return True

    def add(self, to_add: str) -> None:
        self._ensure_feature_branch()
        self.git_cmd(["add", to_add])

    def pull(self, branch: str) -> None:
        click.echo(self.git_cmd(["pull", self.remote_name, branch]).stdout)

    def push(self, branch: str) -> None:
        # you always want to push to origin, even from a fork
        click.echo(self.git_cmd(["push", ORIGIN_NAME, branch]).stdout)

    @staticmethod
    def _parse_reponame_from_url(url: str) -> str:
        without_git_suffix = url.removesuffix(".git")
        if without_git_suffix.startswith("https://") or without_git_suffix.startswith(
            "http://"
        ):
            logger.info(f"Git remote url is in https form {without_git_suffix}")
            split = without_git_suffix.split("/")
            return f"{split[-2]}/{split[-1]}"

        return without_git_suffix.split(":")[-1]

    def full_reponame(self) -> str:
        logger.debug(f"Trying to find {self.remote_name} in {self.remotes}")
        for remote in self.remotes:
            if remote == self.remote_name:
                remote_url = self.git_cmd(
                    ["config", "--get", f"remote.{remote}.url"]
                ).stdout
                logger.debug(
                    f"Remote {remote} found. Parsing its remote url {remote_url}"
                )
                return self._parse_reponame_from_url(remote_url)

        logger.debug("There are no remotes in this repo")
        return ""

    @property
    def git_root(self) -> Path:
        return Path(self.git.git_root)

    def is_branch_merged(self, branch: str) -> bool:
        ret = self.git_cmd(["fetch", self.remote_name, branch])
        return ret.retval != 0 and "couldn't find remote ref" in ret.stderr


class AlpaRepo(LocalRepo):
    def __init__(self, repo_path: Path, gh_api: Optional[GithubAPI] = None) -> None:
        super().__init__(repo_path)

        self.gh_api = gh_api or GithubAPI(self.repo_name)
        self.gh_repo = self.gh_api.get_repo(self.namespace, self.repo_name)

    @abstractmethod
    def create_package(self, package: str) -> None:
        pass

    @abstractmethod
    def request_package(self, package_name: str) -> None:
        pass

    @abstractmethod
    def request_package_delete(self, package: str) -> None:
        pass

    @staticmethod
    def _prepare_cloned_repo(where_to_clone: str, gh_repo: GithubRepo) -> None:
        if not gh_repo.is_fork:
            return

        upstream_clone_url = gh_repo.upstream_clone_url
        assert upstream_clone_url is not None
        subprocess.run(
            ["git", "remote", "add", UPSTREAM_NAME, upstream_clone_url],
            cwd=where_to_clone,
        )

    @staticmethod
    def _get_repo_name_from_url(repo_url: str) -> str:
        return repo_url.split("/")[-1].removesuffix(".git")

    @staticmethod
    def _check_for_permission_and_fork(
        clone_fork: bool, gh_repo: GithubRepo
    ) -> tuple[bool, str]:
        if clone_fork and not gh_repo.is_fork:
            return False, CLONED_REPO_IS_NOT_FORK

        if not clone_fork and gh_repo.is_fork:
            return False, CLONED_REPO_IS_FORK

        if not gh_repo.is_fork and not gh_repo.has_write_access(gh_repo.api_user):
            return False, NO_PERMISSION_FOR_ALPA_REPO

        return True, ""

    @classmethod
    def clone(cls, url: str, clone_fork: bool) -> None:
        # in case of `@` in url -> remove the `git@` prefix form it
        repo_path = urlparse(url.split("@")[-1]).path
        parsed_repo_path = repo_path.strip("/").strip(".git")
        namespace, repo_name = parsed_repo_path.split("/")
        api = GithubAPI(repo_name)
        gh_repo = api.get_repo(namespace, repo_name)

        check_result, stderr = cls._check_for_permission_and_fork(clone_fork, gh_repo)
        if not check_result:
            raise UsageError(stderr)

        cwd = getcwd()
        where_to_clone = f"{cwd}/{cls._get_repo_name_from_url(url)}"
        subprocess.run(["git", "clone", url, where_to_clone], cwd=cwd)
        cls._prepare_cloned_repo(where_to_clone, gh_repo)



"""
These commands need to create LocalRepo -> no GH token required
"""
from os import getcwd
from pathlib import Path

import click
from click import ClickException, Choice
from specfile import Specfile

from alpa.config import MetadataConfig, PackitConfig
from alpa.repository.branch import LocalRepoBranch, AlpaRepoBranch
from alpa.upstream_integration import UpstreamIntegration

pkg_name = click.argument("name", type=str)


@click.command("show-history")
@click.option(
    "-o",
    "--oneline",
    is_flag=True,
    show_default=True,
    default=False,
    help="Each commit is displayed as one line",
)
def show_history(oneline: bool) -> None:
    """Show git history of current package"""
    params = []
    if oneline:
        params.append("--oneline")

    local_repo = LocalRepoBranch(Path(getcwd()))
    click.echo(local_repo.get_history_of_branch(local_repo.branch, params))


@click.command("switch")
@pkg_name
def switch(name: str) -> None:
    """Switch to specified package"""
    LocalRepoBranch(Path(getcwd())).switch_to_package(name)


@click.command("commit")
@click.option(
    "-m",
    "--message",
    type=str,
    default="",
    help="Your commit message not longer than 80 characters.",
)
@click.option("-n", "--no-verify", is_flag=True, help="Do not run pre-commit")
def commit(message: str, no_verify: bool) -> None:
    """Commit your changes in your package's repository"""
    if len(message) > 80:
        raise ClickException("Message longer than 80 characters")

    LocalRepoBranch(Path(getcwd())).commit(message, not no_verify)


@click.command("add")
@click.argument("to_add", type=str, required=True)
def add(to_add: str) -> None:
    """Add files to git history. Basically calls `git add <input>`"""
    LocalRepoBranch(Path(getcwd())).add(to_add)


@click.command("push")
@click.option(
    "-p",
    "--pull-request",
    is_flag=True,
    show_default=True,
    default=False,
    help="This will create pull request on GitHub for you.",
)
def push(pull_request: bool) -> None:
    """Pushes your commited changes to the Alpa repo so you can make PR"""
    repo_path = Path(getcwd())
    local_repo = LocalRepoBranch(repo_path)

    packit_conf = PackitConfig(local_repo.package)
    if not packit_conf.packit_config_file_exists():
        packit_conf.create_packit_config()
        local_repo.git_cmd(["add", ".packit.yaml"])
        local_repo.git_cmd(
            [
                "commit",
                '-m "alpa: automatically add .packit.yaml config to the package"',
            ]
        )

    local_repo.push(local_repo.branch)

    if not pull_request:
        local_repo.git_cmd(["branch", "-d", local_repo.feat_branch])
        return

    alpa = AlpaRepoBranch(repo_path)
    pr = alpa.gh_repo.create_pr(
        title=f"[alpa-cli] Create update of package {local_repo.package}",
        body=(
            "This PR was created automatically via alpa-cli for "
            f"user {alpa.gh_api.gh_user}"
        ),
        source_branch=f"{local_repo.feat_branch}",
        target_branch=local_repo.package,
    )
    click.echo(f"PR#{pr.number} created. URL: {pr.html_url}")


@click.command("pull")
def pull() -> None:
    """Pull last recent changes of package you are on from Alpa repo"""
    local_repo = LocalRepoBranch(Path(getcwd()))
    local_repo.pull(local_repo.branch)


@click.command("list")
@click.option("-p", "--pattern", type=str, default="", help="Optional pattern to match")
def list_(pattern: str) -> None:
    """List all packages or packages matching regex"""
    for pkg in LocalRepoBranch(Path(getcwd())).get_packages(pattern):
        click.echo(pkg)


@click.command("genspec")
@click.option(
    "--lang",
    type=Choice(["python", "java"], case_sensitive=False),
    required=True,
    help="Choose the programming language for which the generator is designed",
)
@click.option(
    "-t",
    "--test",
    default=False,
    help=(
        "Send package with generated spec file to "
        "packit to test whether build will succeed."
    ),
)
def genspec(lang: str, test: bool) -> None:
    """This command uses some existing spec file generators for you"""
    raise NotImplementedError("Not implemented yet (1.0 goal)")


@click.command("create-packit-config")
@click.option(
    "--override",
    is_flag=True,
    show_default=True,
    default=False,
    help="Override packit config file",
)
def create_packit_config(override: bool) -> None:
    """Creates packit config based on metadata.yaml file in package"""
    if not LocalRepoBranch(Path(getcwd())).create_packit_config(override):
        raise ClickException(
            "Packit file already exists. To override it use --override flag."
        )


def _get_chroots_to_build(meta: MetadataConfig, distros: list[str]) -> list[str]:
    chroots = []
    for arch in meta.arch:
        for distro in distros:
            chroots.append(f"{distro}-{arch}")
    return chroots


@click.command("mockbuild")
@click.option(
    "--chroot",
    type=str,
    default="",
    help="Override chroots specified in metadata.yaml by one specific chroot",
)
def mockbuild(chroot: str) -> None:
    """
    Uses mock tool to build package. WARNING: works only on rpm-based systems.

    Builds for all chroots specified in metadata.yaml. Can be overriden by --chroot
     option which does build against one specified chroot.
    """
    meta = MetadataConfig.get_config()
    distros = [chroot] if chroot else list(meta.targets)
    chroots = _get_chroots_to_build(meta, distros)
    UpstreamIntegration(Path(getcwd())).mockbuild(chroots)


@click.command("get-pkg-archive")
def get_pkg_archive() -> None:
    """Gets archive from package config"""
    specfile_path = None
    cwd = Path(getcwd())
    for file in cwd.iterdir():
        if str(file).endswith(".spec"):
            specfile_path = file

    if specfile_path is None:
        raise ClickException(f"No specfile found in {cwd}")

    specfile = Specfile(specfile_path)
    with specfile.sources() as sources:
        source0 = min(sources, key=lambda src: src.number)

    UpstreamIntegration.download_upstream_source(
        source0.expanded_location,
        f"{specfile.expanded_name}-{specfile.expanded_version}",
    )



"""
Class that generates packit.yaml config based on other configs like
Alpa repo or package metadata config
"""


from os import getcwd
from pathlib import Path

from yaml import dump

from alpa.config import AlpaRepoConfig, MetadataConfig

from alpa.constants import PACKIT_CONFIG_NAMES


class PackitConfig:
    def __init__(self, package_name: str) -> None:
        self.package_name = package_name
        self.working_dir = Path(getcwd())
        self.metadata = MetadataConfig.get_config(self.working_dir)
        self.alpa_repo_config = AlpaRepoConfig.get_config()

    def get_packit_config(self) -> dict:
        jobs = [
            {
                "job": "copr_build",
                "trigger": "pull_request",
                "targets": list(self.metadata.targets),
                "owner": self.alpa_repo_config.copr_owner,
                "project": f"{self.alpa_repo_config.copr_repo}-pull-requests",
            },
            {
                "job": "copr_build",
                "trigger": "commit",
                "branch": self.package_name,
                "targets": list(self.metadata.targets),
                "owner": self.alpa_repo_config.copr_owner,
                "project": self.alpa_repo_config.copr_repo,
            },
        ]

        if self.metadata.autoupdate is not None:
            autoupdate_dict = {
                "job": "copr_build",
                "trigger": "commit",
                "branch": f"__alpa_autoupdate_{self.package_name}",
                "targets": list(self.metadata.targets),
                "owner": self.alpa_repo_config.copr_owner,
                "project": f"{self.alpa_repo_config.copr_repo}-pull-requests",
            }
            jobs.append(autoupdate_dict)

        return {
            "specfile_path": f"{self.package_name}.spec",
            "srpm_build_deps": ["pip"],
            "actions": {
                "create-archive": [
                    "pip install pyalpa",
                    'bash -c "alpa get-pkg-archive"',
                    f'bash -c "ls -1 ./{self.package_name}-*.tar.gz"',
                ],
            },
            "jobs": jobs,
        }

    def packit_config_file_exists(self) -> bool:
        for packit_config_name in PACKIT_CONFIG_NAMES:
            files_in_dir = [
                file.name for file in self.working_dir.iterdir() if file.is_file()
            ]
            if packit_config_name in files_in_dir:
                return True

        return False

    def create_packit_config(self) -> None:
        if self.packit_config_file_exists():
            raise FileExistsError("Packit configuration file already exists")

        with open(".packit.yaml", "w") as packit_yaml:
            packit_yaml.write(dump(self.get_packit_config(), sort_keys=False))



from alpa.config.alpa_repo import AlpaRepoConfig
from alpa.config.metadata import MetadataConfig
from alpa.config.packit import PackitConfig


__all__ = [AlpaRepoConfig.__name__, MetadataConfig.__name__, PackitConfig.__name__]



"""
Abstract config class
"""


from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable, Optional, Any

from alpa.exceptions import AlpaConfException


class Config(ABC):
    @staticmethod
    def get_config_file_path(
        path: Path, allowed_file_names: Iterable[str]
    ) -> Optional[Path]:
        for file_name in allowed_file_names:
            full_file_path = path / file_name
            if full_file_path.is_file():
                return full_file_path

        return None

    @staticmethod
    def _check_for_mandatory_key(
        d: dict,
        mandatory_key: str,
        config_file_name: str,
        different_key_name: Optional[str] = None,
    ) -> Any:
        mandatory_val = d.get(mandatory_key)
        if mandatory_val is None:
            missing = (
                mandatory_key if different_key_name is None else different_key_name
            )
            raise AlpaConfException(
                f"The `{missing}` key is missing in {config_file_name}"
            )

        return mandatory_val

    @classmethod
    @abstractmethod
    def get_config(cls) -> "Config":
        pass
